=== modified file 'bzrlib/commands.py'
--- bzrlib/commands.py	2006-06-10 13:27:24 +0000
+++ bzrlib/commands.py	2006-06-18 23:19:23 +0000
@@ -312,6 +312,17 @@
     def name(self):
         return _unsquish_command_name(self.__class__.__name__)
 
+    def plugin_name(self):
+        """Get the name of the plugin that provides this command.
+
+        :return: The name of the plugin or None if the command is builtin.
+        """
+        mod_parts = self.__module__.split('.')
+        if len(mod_parts) >= 3 and mod_parts[1] == 'plugins':
+            return mod_parts[2]
+        else:
+            return None
+
 
 def parse_spec(spec):
     """

=== modified file 'bzrlib/help.py'
--- bzrlib/help.py	2006-05-10 17:48:25 +0000
+++ bzrlib/help.py	2006-06-18 23:19:23 +0000
@@ -85,6 +85,18 @@
     return s
 
 
+def print_command_plugin(cmd_object, outfile, format):
+    """Print the plugin that provides a command object, if any.
+
+    If the cmd_object is provided by a plugin, prints the plugin name to
+    outfile using the provided format string.
+    """
+    plugin_name = cmd_object.plugin_name()
+    if plugin_name is not None:
+        out_str = '(From plugin "%s")' % plugin_name
+        outfile.write(format % out_str)
+
+
 def help_on_command(cmdname, outfile=None):
     from bzrlib.commands import get_cmd_object
 
@@ -107,6 +119,8 @@
 
     print >>outfile
 
+    print_command_plugin(cmd_object, outfile, '%s\n\n')
+
     outfile.write(doc)
     if doc[-1] != '\n':
         outfile.write('\n')
@@ -156,6 +170,10 @@
         if cmd_object.hidden:
             continue
         print >>outfile, command_usage(cmd_object)
+
+        plugin_name = cmd_object.plugin_name()
+        print_command_plugin(cmd_object, outfile, '        %s\n')
+
         cmd_help = cmd_object.help()
         if cmd_help:
             firstline = cmd_help.split('\n', 1)[0]

=== modified file 'bzrlib/tests/test_plugins.py'
--- bzrlib/tests/test_plugins.py	2005-12-16 03:43:14 +0000
+++ bzrlib/tests/test_plugins.py	2006-06-18 23:19:23 +0000
@@ -22,9 +22,12 @@
 # comments.
 
 import os
+from StringIO import StringIO
 
 import bzrlib.plugin
 import bzrlib.plugins
+import bzrlib.commands
+import bzrlib.help
 from bzrlib.tests import TestCaseInTempDir
 from bzrlib.osutils import pathjoin, abspath
 
@@ -137,3 +140,57 @@
             if getattr(bzrlib.plugins, 'plugin', None):
                 del bzrlib.plugins.plugin
         self.failIf(getattr(bzrlib.plugins, 'plugin', None))
+
+
+class TestPluginHelp(TestCaseInTempDir):
+
+    def split_help_commands(self):
+        help = {}
+        current = None
+        for line in self.capture('help commands').splitlines():
+            if line.startswith('bzr '):
+                current = line.split()[1]
+            help[current] = help.get(current, '') + line
+
+        return help
+
+    def test_plugin_help_builtins_unaffected(self):
+        # Check we don't get false positives
+        help_commands = self.split_help_commands()
+        for cmd_name in bzrlib.commands.builtin_command_names():
+            if cmd_name in bzrlib.commands.plugin_command_names():
+                continue
+            help = StringIO()
+            try:
+                bzrlib.help.help_on_command(cmd_name, help)
+            except NotImplementedError:
+                # some commands have no help
+                pass
+            else:
+                help.seek(0)
+                self.assertNotContainsRe(help.read(), 'From plugin "[^"]*"')
+
+            if help in help_commands.keys():
+                # some commands are hidden
+                help = help_commands[cmd_name]
+                self.assertNotContainsRe(help, 'From plugin "[^"]*"')
+
+    def test_plugin_help_shows_plugin(self):
+        # Create a test plugin
+        os.mkdir('plugin_test')
+        f = open(pathjoin('plugin_test', 'myplug.py'), 'w')
+        f.write(PLUGIN_TEXT)
+        f.close()
+
+        try:
+            # Check its help
+            bzrlib.plugin.load_from_dirs(['plugin_test'])
+            bzrlib.commands.register_command( bzrlib.plugins.myplug.cmd_myplug)
+            help = self.capture('help myplug')
+            self.assertContainsRe(help, 'From plugin "myplug"')
+            help = self.split_help_commands()['myplug']
+            self.assertContainsRe(help, 'From plugin "myplug"')
+        finally:
+            # remove the plugin 'plugin'
+            if getattr(bzrlib.plugins, 'plugin', None):
+                del bzrlib.plugins.plugin

